# Std lib
from typing import List, Union

# 3rd party lib
import numpy as np
from numpy import array
from numpy.typing import NDArray
from shapely import buffer, Polygon, MultiPolygon, LineString, prepare

# Local lib
from ..beams import BeamType
from ..mill import Mill


def interpolate_line(line: LineString, distances: np.ndarray) -> np.ndarray:
    coords = np.array(line.coords)
    if line.is_closed:
        coords = np.vstack([coords, coords[0]])

    dxdy = np.diff(coords, axis=0)
    seg_lengths = np.hypot(dxdy[:, 0], dxdy[:, 1], dtype=np.float32)
    cum_lengths = np.cumsum(seg_lengths)
    total_length = cum_lengths[-1] if len(cum_lengths) > 0 else 0.0

    # Vectorized computation
    indices = np.searchsorted(cum_lengths, distances, side='right')
    valid = (indices > 0) & (indices <= len(seg_lengths))
    prev_lengths = np.where(indices > 0, cum_lengths[np.clip(indices - 1, 0, None)], 0)
    t = (distances - prev_lengths) / np.where(seg_lengths[indices] == 0, 1, seg_lengths[indices])
    t = np.clip(t, 0, 1)

    x = coords[indices, 0] + t * (coords[indices + 1, 0] - coords[indices, 0])
    y = coords[indices, 1] + t * (coords[indices + 1, 1] - coords[indices, 1])
    return np.column_stack((x, y))


def contour_inward(shape: Union[Polygon, MultiPolygon], mill: Mill, beam: BeamType,
                   rasterize: bool = True) -> List[NDArray]:
    beam_radius = beam.get_radius(units="nm")
    overlap = mill.get_overlap()
    step = 2 * beam_radius * (1 - overlap)
    dwelltime = int(mill.get_dwell_time("ns")) / 100

    contours = []
    prepare(shape)  # Precompute spatial indexes for faster operations

    def add_linestrings(geom):
        if isinstance(geom, Polygon):
            if not geom.exterior.is_empty:
                contours.append(geom.exterior)
            contours.extend(geom.interiors)
        elif isinstance(geom, MultiPolygon):
            for poly in geom.geoms:
                if not poly.exterior.is_empty:
                    contours.append(poly.exterior)
                contours.extend(poly.interiors)

    buffered = buffer(shape, -beam_radius, join_style="mitre")
    quad_segs = 32  # Reduced from 100 for initial buffer
    k = 0

    while True:
        if k != 0:
            quad_segs = max(8, int(quad_segs * 0.8))  # Decrease resolution inward
            buffered = buffer(buffered, -step, join_style="round", quad_segs=quad_segs)
            if buffered.is_empty:
                break
        add_linestrings(buffered)
        k += 1

    streamlines = []
    if not rasterize:
        for line in contours:
            coords = np.array(line.coords)
            streamline = np.column_stack((
                np.zeros(len(coords)),
                coords[:, 0],
                coords[:, 1],
                np.zeros(len(coords))
            ))
            streamlines.append(streamline)
        return streamlines

    for line in contours:
        line_length = line.length
        nr_steps = int(np.rint(line_length / step))
        points = []

        if line.is_closed and nr_steps == 3:
            if line_length / 2 > beam_radius * 1.5:
                distances = np.linspace(0, line_length, nr_steps + 1, endpoint=False)
                points = interpolate_line(line, distances)
                centroid = line.centroid
                if line.distance(centroid) > beam_radius:
                    points = np.vstack([points, [centroid.x, centroid.y]])
            else:
                points = np.array([[line.centroid.x, line.centroid.y]])
        elif not line.is_closed and nr_steps == 2:
            points = np.array([[line.centroid.x, line.centroid.y]])
        else:
            distances = np.linspace(0, line_length, nr_steps, endpoint=not line.is_closed)
            points = interpolate_line(line, distances)

        if points.size == 0:
            continue

        codes = np.full(len(points), 2, dtype=int)
        codes[0] = 1
        dwelltimes = np.full(len(points), dwelltime)
        streamline = np.column_stack((dwelltimes, points[:, 0], points[:, 1], codes))

        if line.is_closed:
            closing = np.array([[1, streamline[0, 1], streamline[0, 2], 0]])
        else:
            closing = np.array([[1, streamline[-1, 1], streamline[-1, 2], 0]])
        streamlines.append(np.vstack([streamline, closing]))

    return streamlines


def contour_outward(shape: Polygon | MultiPolygon, mill: Mill, beam: BeamType, rasterize: bool = True) -> List[NDArray]:
    return contour_inward(shape, mill, beam, rasterize)[::-1]