from __future__ import annotations

# Std lib
from typing import Unpack, TypedDict

# 3rd party lib
from matplotlib.font_manager import FontProperties
from matplotlib.textpath import TextPath
from shapely import Polygon, MultiPolygon, unary_union
from shapely.affinity import translate

# Local lib
from .shape import Shape


class TextKwargs(TypedDict, total=False):
    x: float
    y: float
    fontsize: int
    millable: bool


class Text(Shape):

    # region Class Body

    # Class variables
    _default_geometry = Shape._default_geometry.copy()
    _default_geometry.update({"fontsize": 1000})

    # Instance variables
    _fontsize: int
    _text: str

    __slots__ = ["_fontsize", "_text"]

    # endregion Class Body

    # region Dev Methods

    def __init__(self, text: str, **kwargs: Unpack[TextKwargs]) -> None:
        super().__init__(**kwargs)
        self._text = text

    def _get_polygon(self) -> MultiPolygon:
        """
        Converts a given text string into a Shapely MultiPolygon and centers it
        by shifting it negatively by half its width and half its height.
        """
        # Set up font properties
        font_prop = FontProperties()

        # Create the text path at origin (0, 0)
        path = TextPath(self._get_absolute_position(), self._text, size=self._fontsize * (1 / 0.725), prop=font_prop)

        # Extract polygons from the path
        polygons = []
        for vertices in path.to_polygons():
            if len(vertices) >= 3:  # Ensure polygon has at least 3 points
                polygons.append(Polygon(vertices))

        # If no valid polygons, return an empty MultiPolygon
        if not polygons:
            return MultiPolygon()

        # Union polygons to handle overlaps and holes
        union = unary_union(polygons)

        # Convert single polygon to MultiPolygon if necessary
        if union.geom_type == 'Polygon':
            union = MultiPolygon([union])

        # Get the bounding box (minx, miny, maxx, maxy)
        minx, miny, maxx, maxy = union.bounds

        # Compute shift values
        shift_x = (maxx - minx) / 2
        shift_y = (maxy - miny) / 2

        # Apply translation to shift the geometry
        centered_geometry = translate(union, xoff=-shift_x, yoff=-shift_y)

        return centered_geometry

    # endregion Dev Methods

    # region User Methods

    def copy(self, **kwargs) -> Text:
        return super().copy(**kwargs)

    # endregion User Methods

    # region Properties

    @property
    def text(self) -> str:
        return self._text

    @text.setter
    def text(self, text: str):
        self._text = str(text)

    @property
    def font_size(self) -> int:
        return self._fontsize

    @font_size.setter
    def font_size(self, value: int):
        self._fontsize = int(value)

    # endregion Properties
